from functools import partial
import warnings

import numpy as np
import pandas as pd
import bioframe

from .lib.numutils import LazyToeplitz
import warnings


def make_bin_aligned_windows(
    binsize, chroms, centers_bp, flank_bp=0, region_start_bp=0, ignore_index=False
):
    """
    Convert genomic loci into bin spans on a fixed bin-size segmentation of a
    genomic region. Window limits are adjusted to align with bin edges.

    Parameters
    -----------
    binsize : int
        Bin size (resolution) in base pairs.
    chroms : 1D array-like
        Column of chromosome names.
    centers_bp : 1D or nx2 array-like
        If 1D, center points of each window. If 2D, the starts and ends.
    flank_bp : int
        Distance in base pairs to extend windows on either side.
    region_start_bp : int, optional
        If region is a subset of a chromosome, shift coordinates by this amount.
        Default is 0.

    Returns
    -------
    DataFrame with columns:
        'chrom'        - chromosome
        'start', 'end' - window limits in base pairs
        'lo', 'hi'     - window limits in bins

    """
    if not (flank_bp % binsize == 0):
        raise ValueError("Flanking distance must be divisible by the bin size.")

    if isinstance(chroms, pd.Series) and not ignore_index:
        index = chroms.index
    else:
        index = None

    chroms = np.asarray(chroms)
    centers_bp = np.asarray(centers_bp)
    if len(centers_bp.shape) == 2:
        left_bp = centers_bp[:, 0]
        right_bp = centers_bp[:, 1]
    else:
        left_bp = right_bp = centers_bp

    if np.any(left_bp > right_bp):
        raise ValueError("Found interval with end > start.")

    left = left_bp - region_start_bp
    right = right_bp - region_start_bp
    left_bin = (left / binsize).astype(int)
    right_bin = (right / binsize).astype(int)
    flank_bin = flank_bp // binsize

    lo = left_bin - flank_bin
    hi = right_bin + flank_bin + 1
    windows = pd.DataFrame(index=index)
    windows["chrom"] = chroms
    windows["start"] = lo * binsize
    windows["end"] = hi * binsize
    windows["lo"] = lo
    windows["hi"] = hi
    return windows


def assign_regions(features, supports):
    """
    For each feature in features dataframe assign the genomic region (support)
    that overlaps with it. In case if feature overlaps multiple supports, the 
    region with largest overlap will be reported.
    """

    index_name = features.index.name  # Store the name of index
    features = (
        features.copy().reset_index()
    )  # Store the original features' order as a column with original index

    if "chrom" in features.columns:
        overlap = bioframe.overlap(
            features,
            supports,
            how="left",
            cols1=["chrom", "start", "end"],
            cols2=["chrom", "start", "end"],
            keep_order=True,
            return_overlap=True,
        )
        overlap_columns = overlap.columns  # To filter out duplicates later
        overlap["overlap_length"] = overlap["overlap_end"] - overlap["overlap_start"]
        # Filter out overlaps with multiple regions:
        overlap = (
            overlap.sort_values("overlap_length", ascending=False)
            .drop_duplicates(overlap_columns, keep="first")
            .sort_index()
        )
        # Copy single column with overlapping region name:
        features["region"] = overlap["name_2"]

    if "chrom1" in features.columns:
        for idx in ("1", "2"):
            overlap = bioframe.overlap(
                features,
                supports,
                how="left",
                cols1=[f"chrom{idx}", f"start{idx}", f"end{idx}"],
                cols2=[f"chrom", f"start", f"end"],
                keep_order=True,
                return_overlap=True,
            )
            overlap_columns = overlap.columns  # To filter out duplicates later
            overlap[f"overlap_length{idx}"] = (
                overlap[f"overlap_end{idx}"] - overlap[f"overlap_start{idx}"]
            )
            # Filter out overlaps with multiple regions:
            overlap = (
                overlap.sort_values(f"overlap_length{idx}", ascending=False)
                .drop_duplicates(overlap_columns, keep="first")
                .sort_index()
            )
            # Copy single column with overlapping region name:
            features[f"region{idx}"] = overlap["name_2"]

        # Form a single column with region names where region1 == region2, and np.nan in other cases:
        features["region"] = np.where(
            features["region1"] == features["region2"], features["region1"], np.nan
        )
        features = features.drop(
            ["region1", "region2"], axis=1
        )  # Remove unnecessary columns

    features = features.set_index(
        index_name if not index_name is None else "index"
    )  # Restore the original index
    features.index.name = index_name  # Restore original index title
    return features


def _pileup(data_select, data_snip, arg):
    support, feature_group = arg

    # return empty snippets if region is unannotated:
    if len(support) == 0:
        if "start" in feature_group:  # on-diagonal off-region case:
            lo = feature_group["lo"].values
            hi = feature_group["hi"].values
            s = hi - lo  # Shape of individual snips
            assert (
                s.max() == s.min()
            ), "Pileup accepts only the windows of the same size"
            stack = np.full((s[0], s[0], len(feature_group)), np.nan)
        else:  # off-diagonal off-region case:
            lo1 = feature_group["lo1"].values
            hi1 = feature_group["hi1"].values
            lo2 = feature_group["lo2"].values
            hi2 = feature_group["hi2"].values
            s1 = hi1 - lo1  # Shape of individual snips
            s2 = hi1 - lo1
            assert (
                s1.max() == s1.min()
            ), "Pileup accepts only the windows of the same size"
            assert (
                s2.max() == s2.min()
            ), "Pileup accepts only the windows of the same size"
            stack = np.full((s1[0], s2[0], len(feature_group)), np.nan)

        return stack, feature_group["_rank"].values

    # check if support region is on- or off-diagonal
    if len(support) == 2:
        region1, region2 = support
    else:
        region1 = region2 = support

    # check if features are on- or off-diagonal
    if "start" in feature_group:
        s1 = feature_group["start"].values
        e1 = feature_group["end"].values
        s2, e2 = s1, e1
    else:
        s1 = feature_group["start1"].values
        e1 = feature_group["end1"].values
        s2 = feature_group["start2"].values
        e2 = feature_group["end2"].values

    data = data_select(region1, region2)
    stack = list(map(partial(data_snip, data, region1, region2), zip(s1, e1, s2, e2)))

    return np.dstack(stack), feature_group["_rank"].values


def pileup(features, data_select, data_snip, map=map):
    """
    Handles on-diagonal and off-diagonal cases.

    Parameters
    ----------
    features : DataFrame
        Table of features. Requires columns ['chrom', 'start', 'end'].
        Or ['chrom1', 'start1', 'end1', 'chrom1', 'start2', 'end2'].
        start, end are bp coordinates.
        lo, hi are bin coordinates.

    data_select : callable
        Callable that takes a region as argument and returns
        the data, mask and bin offset of a support region

    data_snip : callable
        Callable that takes data, mask and a 2D bin span (lo1, hi1, lo2, hi2)
        and returns a snippet from the selected support region


    """
    if features.region.isnull().any():
        warnings.warn(
            "Some features do not have regions assigned! Some snips will be empty."
        )

    features = features.copy()
    features["region"] = features.region.fillna(
        ""
    )  # fill in unanotated regions with empty string
    features["_rank"] = range(len(features))

    # cumul_stack = []
    # orig_rank = []
    cumul_stack, orig_rank = zip(
        *map(
            partial(_pileup, data_select, data_snip),
            # Note that unannotated regions will form a separate group
            features.groupby("region", sort=False),
        )
    )

    # Restore the original rank of the input features
    cumul_stack = np.dstack(cumul_stack)
    orig_rank = np.concatenate(orig_rank)

    idx = np.argsort(orig_rank)
    cumul_stack = cumul_stack[:, :, idx]
    return cumul_stack


def pair_sites(sites, separation, slop):
    """
    Create "hand" intervals to the right and to the left of each site.
    Then join right hands with left hands to pair sites together.

    """
    from bioframe.tools import tsv, bedtools

    mids = (sites["start"] + sites["end"]) // 2
    left_hand = sites[["chrom"]].copy()
    left_hand["start"] = mids - separation - slop
    left_hand["end"] = mids - separation + slop
    left_hand["site_id"] = left_hand.index
    left_hand["direction"] = "L"
    left_hand["snip_mid"] = mids
    left_hand["snip_strand"] = sites["strand"]

    right_hand = sites[["chrom"]].copy()
    right_hand["start"] = mids + separation - slop
    right_hand["end"] = mids + separation + slop
    right_hand["site_id"] = right_hand.index
    right_hand["direction"] = "R"
    right_hand["snip_mid"] = mids
    right_hand["snip_strand"] = sites["strand"]

    # ignore out-of-bounds hands
    mask = (left_hand["start"] > 0) & (right_hand["start"] > 0)
    left_hand = left_hand[mask].copy()
    right_hand = right_hand[mask].copy()

    # intersect right hands (left anchor site)
    # with left hands (right anchor site)
    with tsv(right_hand) as R, tsv(left_hand) as L:
        out = bedtools.intersect(a=R.name, b=L.name, wa=True, wb=True)
        out.columns = [c + "_r" for c in right_hand.columns] + [
            c + "_l" for c in left_hand.columns
        ]
    return out


class CoolerSnipper:
    def __init__(self, clr, cooler_opts=None, regions=None):

        if regions is None:
            regions = pd.DataFrame(
                [(chrom, 0, l, chrom) for chrom, l in clr.chromsizes.items()],
                columns=["chrom", "start", "end", "name"],
            )
        self.regions = regions.set_index("name")

        self.clr = clr
        self.binsize = self.clr.binsize
        self.offsets = {}
        self.pad = True
        self.cooler_opts = {} if cooler_opts is None else cooler_opts
        self.cooler_opts.setdefault("sparse", True)

    def select(self, region1, region2):
        region1_coords = self.regions.loc[region1]
        region2_coords = self.regions.loc[region2]
        self.offsets[region1] = self.clr.offset(region1_coords) - self.clr.offset(
            region1_coords[0]
        )
        self.offsets[region2] = self.clr.offset(region2_coords) - self.clr.offset(
            region2_coords[0]
        )
        matrix = self.clr.matrix(**self.cooler_opts).fetch(
            region1_coords, region2_coords
        )
        self._isnan1 = np.isnan(self.clr.bins()["weight"].fetch(region1_coords).values)
        self._isnan2 = np.isnan(self.clr.bins()["weight"].fetch(region2_coords).values)
        if self.cooler_opts["sparse"]:
            matrix = matrix.tocsr()
        return matrix

    def snip(self, matrix, region1, region2, tup):
        s1, e1, s2, e2 = tup
        offset1 = self.offsets[region1]
        offset2 = self.offsets[region2]
        binsize = self.binsize
        lo1, hi1 = (s1 // binsize) - offset1, (e1 // binsize) - offset1
        lo2, hi2 = (s2 // binsize) - offset2, (e2 // binsize) - offset2
        assert hi1 >= 0
        assert hi2 >= 0

        m, n = matrix.shape
        dm, dn = hi1 - lo1, hi2 - lo2
        out_of_bounds = False
        pad_left = pad_right = pad_bottom = pad_top = None
        if lo1 < 0:
            pad_bottom = -lo1
            out_of_bounds = True
        if lo2 < 0:
            pad_left = -lo2
            out_of_bounds = True
        if hi1 > m:
            pad_top = dm - (hi1 - m)
            out_of_bounds = True
        if hi2 > n:
            pad_right = dn - (hi2 - n)
            out_of_bounds = True

        if out_of_bounds:
            i0 = max(lo1, 0)
            i1 = min(hi1, m)
            j0 = max(lo2, 0)
            j1 = min(hi2, n)
            snippet = np.full((dm, dn), np.nan)
        #             snippet[pad_bottom:pad_top,
        #                     pad_left:pad_right] = matrix[i0:i1, j0:j1].toarray()
        else:
            snippet = matrix[lo1:hi1, lo2:hi2].toarray().astype("float")
            snippet[self._isnan1[lo1:hi1], :] = np.nan
            snippet[:, self._isnan2[lo2:hi2]] = np.nan
        return snippet


class ObsExpSnipper:
    def __init__(self, clr, expected, cooler_opts=None, regions=None):
        self.clr = clr
        self.expected = expected

        # Detecting the columns for the detection of regions
        columns = expected.columns
        assert len(columns) > 0
        if "region" not in columns:
            if "chrom" in columns:
                self.expected = self.expected.rename(columns={"chrom": "region"})
                warnings.warn(
                    "The expected dataframe appears to be in the old format."
                    "It should have a `region` column, not `chrom`."
                )
            else:
                raise ValueError(
                    "Please check the expected dataframe, it has no `region` column"
                )
        if regions is None:
            if set(self.expected["region"]).issubset(clr.chromnames):
                regions = pd.DataFrame(
                    [(chrom, 0, l, chrom) for chrom, l in clr.chromsizes.items()],
                    columns=["chrom", "start", "end", "name"],
                )
            else:
                raise ValueError(
                    "Please provide the regions table, if region names"
                    "are not simply chromosome names."
                )
        self.regions = regions.set_index("name")

        try:
            for region_name, group in self.expected.groupby("region"):
                n_diags = group.shape[0]
                region = self.regions.loc[region_name]
                lo, hi = self.clr.extent(region)
                assert n_diags == (hi - lo)
        except AssertionError:
            raise ValueError(
                "Region shape mismatch between expected and cooler. "
                "Are they using the same resolution?"
            )

        self.binsize = self.clr.binsize
        self.offsets = {}
        self.pad = True
        self.cooler_opts = {} if cooler_opts is None else cooler_opts
        self.cooler_opts.setdefault("sparse", True)

    def select(self, region1, region2):
        assert region1 == region2, "ObsExpSnipper is implemented for cis contacts only."
        region1_coords = self.regions.loc[region1]
        region2_coords = self.regions.loc[region2]
        self.offsets[region1] = self.clr.offset(region1_coords) - self.clr.offset(
            region1_coords[0]
        )
        self.offsets[region2] = self.clr.offset(region2_coords) - self.clr.offset(
            region2_coords[0]
        )
        matrix = self.clr.matrix(**self.cooler_opts).fetch(
            region1_coords, region2_coords
        )
        if self.cooler_opts["sparse"]:
            matrix = matrix.tocsr()
        self._isnan1 = np.isnan(self.clr.bins()["weight"].fetch(region1_coords).values)
        self._isnan2 = np.isnan(self.clr.bins()["weight"].fetch(region2_coords).values)
        self._expected = LazyToeplitz(
            self.expected.groupby("region").get_group(region1)["balanced.avg"].values
        )
        return matrix

    def snip(self, matrix, region1, region2, tup):
        s1, e1, s2, e2 = tup
        offset1 = self.offsets[region1]
        offset2 = self.offsets[region2]
        binsize = self.binsize
        lo1, hi1 = (s1 // binsize) - offset1, (e1 // binsize) - offset1
        lo2, hi2 = (s2 // binsize) - offset2, (e2 // binsize) - offset2
        assert hi1 >= 0
        assert hi2 >= 0

        m, n = matrix.shape
        dm, dn = hi1 - lo1, hi2 - lo2
        out_of_bounds = False
        pad_left = pad_right = pad_bottom = pad_top = None
        if lo1 < 0:
            pad_bottom = -lo1
            out_of_bounds = True
        if lo2 < 0:
            pad_left = -lo2
            out_of_bounds = True
        if hi1 > m:
            pad_top = dm - (hi1 - m)
            out_of_bounds = True
        if hi2 > n:
            pad_right = dn - (hi2 - n)
            out_of_bounds = True

        if out_of_bounds:
            i0 = max(lo1, 0)
            i1 = min(hi1, m)
            j0 = max(lo2, 0)
            j1 = min(hi2, n)
            return np.full((dm, dn), np.nan)
        #             snippet[pad_bottom:pad_top,
        #                     pad_left:pad_right] = matrix[i0:i1, j0:j1].toarray()
        else:
            snippet = matrix[lo1:hi1, lo2:hi2].toarray()
            snippet[self._isnan1[lo1:hi1], :] = np.nan
            snippet[:, self._isnan2[lo2:hi2]] = np.nan

        e = self._expected[lo1:hi1, lo2:hi2]
        return snippet / e


class ExpectedSnipper:
    def __init__(self, clr, expected, regions=None):
        self.clr = clr
        self.expected = expected
        # Detecting the columns for the detection of regions
        columns = expected.columns
        assert len(columns) > 0
        if "region" not in columns:
            if "chrom" in columns:
                self.expected = self.expected.rename(columns={"chrom": "region"})
                warnings.warn(
                    "The expected dataframe appears to be in the old format."
                    "It should have a `region` column, not `chrom`."
                )
            else:
                raise ValueError(
                    "Please check the expected dataframe, it has no `region` column"
                )
        if regions is None:
            if set(self.expected["region"]).issubset(clr.chromnames):
                regions = pd.DataFrame(
                    [(chrom, 0, l, chrom) for chrom, l in clr.chromsizes.items()],
                    columns=["chrom", "start", "end", "name"],
                )
            else:
                raise ValueError(
                    "Please provide the regions table, if region names"
                    "are not simply chromosome names."
                )
        self.regions = regions.set_index("name")

        try:
            for region_name, group in self.expected.groupby("region"):
                n_diags = group.shape[0]
                region = self.regions.loc[region_name]
                lo, hi = self.clr.extent(region)
                assert n_diags == (hi - lo)
        except AssertionError:
            raise ValueError(
                "Region shape mismatch between expected and cooler. "
                "Are they using the same resolution?"
            )

        self.binsize = self.clr.binsize
        self.offsets = {}

    def select(self, region1, region2):
        assert (
            region1 == region2
        ), "ExpectedSnipper is implemented for cis contacts only."
        region1_coords = self.regions.loc[region1]
        region2_coords = self.regions.loc[region2]
        self.offsets[region1] = self.clr.offset(region1_coords) - self.clr.offset(
            region1_coords[0]
        )
        self.offsets[region2] = self.clr.offset(region2_coords) - self.clr.offset(
            region2_coords[0]
        )
        self.m = np.diff(self.clr.extent(region1_coords))
        self.n = np.diff(self.clr.extent(region2_coords))
        self._expected = LazyToeplitz(
            self.expected.groupby("region").get_group(region1)["balanced.avg"].values
        )
        return self._expected

    def snip(self, exp, region1, region2, tup):
        s1, e1, s2, e2 = tup
        offset1 = self.offsets[region1]
        offset2 = self.offsets[region2]
        binsize = self.binsize
        lo1, hi1 = (s1 // binsize) - offset1, (e1 // binsize) - offset1
        lo2, hi2 = (s2 // binsize) - offset2, (e2 // binsize) - offset2
        assert hi1 >= 0
        assert hi2 >= 0
        dm, dn = hi1 - lo1, hi2 - lo2

        if lo1 < 0 or lo2 < 0 or hi1 > self.m or hi2 > self.n:
            return np.full((dm, dn), np.nan)

        snippet = exp[lo1:hi1, lo2:hi2]
        return snippet



# create a test for the chunking versions of 'get_adjusted_expected_tile_some_nans':

import numpy as np
import pandas as pd

import os.path as op

from cooltools import dotfinder
from cooltools.lib.numutils import LazyToeplitz


# adjust the path for data:
testdir = op.realpath(op.dirname(__file__))

# mock input data location:
mock_input = op.join(testdir, "data", "mock_inputs.npz")
mock_result = op.join(testdir, "data", "mock_res.csv.gz")


# load bunch of array from a numpy npz container:
arrays_loaded = np.load(mock_input)
# snippets of M_raw, M_ice, E_ice and v_ice are supposed
# to be there ...
mock_M_raw = arrays_loaded["mock_M_raw"]
mock_M_ice = arrays_loaded["mock_M_ice"]
mock_E_ice = arrays_loaded["mock_E_ice"]
mock_v_ice = arrays_loaded["mock_v_ice"]

# 1D expected extracted for tiling-tests:
mock_exp = LazyToeplitz(mock_E_ice[0, :])

# we need w-edge for tiling procedures:
w = 3
# p = 1
# kernel type: 'donut'
# # just a simple donut kernel for testing:
kernel = np.array(
    [
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 0, 0, 0, 1, 1],
        [0, 0, 0, 0, 0, 0, 0],
        [1, 1, 0, 0, 0, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
    ]
)
# bin size:
# mock data were extracted
# from 20kb matrix:
b = 20000
# 2MB aroud diagonal to look at:
band = 2e6
# 1MB aroud diagonal to look at:
band_1 = 1e6

# start, stop for tiling procedures:
start, stop = 0, len(mock_M_raw)

# load mock results:
mock_res = (
    pd.read_csv(mock_result)
    .rename(columns={"row": "bin1_id", "col": "bin2_id"})
)


def test_adjusted_expected_tile_some_nans_and_diag_tiling():
    print("Running tile some nans la_exp test + diag tiling")
    # first, generate that locally-adjusted expected:
    nnans = 1
    band_1_idx = int(band_1 / b)
    res_df = pd.DataFrame([])
    for tile in dotfinder.diagonal_matrix_tiling(
        start, stop, bandwidth=band_1_idx, edge=w
    ):
        # let's keep i,j-part explicit here:
        tilei, tilej = tile, tile
        # define origin:
        origin = (tilei[0], tilej[0])
        # RAW observed matrix slice:
        observed = mock_M_raw[slice(*tilei), slice(*tilej)]
        # trying new expected function:
        expected = mock_exp[slice(*tilei), slice(*tilej)]
        # for diagonal chuynking/tiling tilei==tilej:
        ice_weight = mock_v_ice[slice(*tilei)]
        # that's the main working function from dotfinder:
        res = dotfinder.get_adjusted_expected_tile_some_nans(
            origin=origin,
            observed=observed,
            expected=expected,
            bal_weights=ice_weight,
            kernels={"donut": kernel, "footprint": np.ones_like(kernel)},
        )
        is_inside_band = res["bin1_id"] > (res["bin2_id"] - band_1_idx)
        # new style, selecting good guys:
        does_comply_nans = res["la_exp." + "footprint" + ".nnans"] < nnans
        # so, selecting inside band and nNaNs compliant results and append:
        res_df = res_df.append(
            res[is_inside_band & does_comply_nans], ignore_index=True
        )

    # drop dups (from overlaping tiles), sort and reset index:
    res_df = (
        res_df
        .drop_duplicates()
        .sort_values(by=["bin1_id", "bin2_id"])
        .reset_index(drop=True)
    )

    # prepare mock_data for comparison:
    # get a subset of mock results (inside 1Mb band):
    is_inside_band_1 = mock_res["bin1_id"] > (mock_res["bin2_id"] - band_1_idx)
    mock_res_1 = mock_res[is_inside_band_1].reset_index(drop=True)
    # apparently sorting is needed in this case:
    mock_res_1 = (
        mock_res_1
        .drop_duplicates()
        .sort_values(by=["bin1_id", "bin2_id"])
        .reset_index(drop=True)
    )

    # ACTUAL TESTS:
    # integer part of DataFrame must equals exactly:
    assert res_df[["bin1_id", "bin2_id"]].equals(mock_res_1[["bin1_id", "bin2_id"]])
    # compare floating point part separately:
    assert np.isclose(
        res_df["la_exp." + "donut" + ".value"],
        mock_res_1["la_expected"],
        equal_nan=True,
    ).all()


def test_adjusted_expected_tile_some_nans_and_square_tiling():
    print("Running tile some nans la_exp test + square tiling")
    # first, generate that locally-adjusted expected:
    nnans = 1
    band_idx = int(band / b)
    res_df = pd.DataFrame([])
    for tilei, tilej in dotfinder.square_matrix_tiling(
        start, stop, step=40, edge=w, square=False
    ):
        # define origin:
        origin = (tilei[0], tilej[0])
        # RAW observed matrix slice:
        observed = mock_M_raw[slice(*tilei), slice(*tilej)]
        # trying new expected function:
        expected = mock_exp[slice(*tilei), slice(*tilej)]
        # for diagonal chuynking/tiling tilei==tilej:
        ice_weight_i = mock_v_ice[slice(*tilei)]
        ice_weight_j = mock_v_ice[slice(*tilej)]
        # that's the main working function from dotfinder:
        res = dotfinder.get_adjusted_expected_tile_some_nans(
            origin=origin,
            observed=observed,
            expected=expected,
            bal_weights=(ice_weight_i, ice_weight_j),
            kernels={"donut": kernel, "footprint": np.ones_like(kernel)},
            # nan_threshold=1,
            verbose=False,
        )
        is_inside_band = res["bin1_id"] > (res["bin2_id"] - band_idx)
        # new style, selecting good guys:
        does_comply_nans = res["la_exp." + "footprint" + ".nnans"] < nnans
        # so, select inside band and nNaNs compliant results and append:
        res_df = res_df.append(
            res[is_inside_band & does_comply_nans], ignore_index=True
        )

    # drop dups (from overlaping tiles), sort and reset index:
    res_df = (
        res_df
        .drop_duplicates()
        .sort_values(by=["bin1_id", "bin2_id"])
        .reset_index(drop=True)
    )

    # prepare mock_data for comparison:
    # apparently sorting is needed in this case:
    mock_res_sorted = (
        mock_res
        .drop_duplicates()
        .sort_values(by=["bin1_id", "bin2_id"])
        .reset_index(drop=True)
    )

    # ACTUAL TESTS:
    # integer part of DataFrame must equals exactly:
    assert res_df[["bin1_id", "bin2_id"]].equals(
        mock_res_sorted[["bin1_id", "bin2_id"]]
    )
    # compare floating point part separately:
    assert np.isclose(
        res_df["la_exp." + "donut" + ".value"],
        mock_res_sorted["la_expected"],
        equal_nan=True,
    ).all()


def test_adjusted_expected_tile_some_nans_and_square_tiling_diag_band():
    print("Running tile some nans la_exp test + square tiling + diag band")
    # # Essentially, testing this function:
    # def chrom_chunk_generator_s(chroms, w, band):
    #     for chrom in chroms:
    #         chr_start, chr_stop = the_c.extent(chrom)
    #         for tilei, tilej in square_matrix_tiling(chr_start, chr_stop, step, w):
    #             # check if a given tile intersects with
    #             # with the diagonal band of interest ...
    #             diag_from = tilej[0] - tilei[1]
    #             diag_to   = tilej[1] - tilei[0]
    #             #
    #             band_from = 0
    #             band_to   = band
    #             # we are using this >2w trick to exclude
    #             # tiles from the lower triangle from calculations ...
    #             if (min(band_to,diag_to) - max(band_from,diag_from)) > 2*w:
    #                 yield chrom, tilei, tilej
    # first, generate that locally-adjusted expected:
    nnans = 1
    band_idx = int(band / b)
    res_df = pd.DataFrame([])
    for tilei, tilej in dotfinder.square_matrix_tiling(
        start, stop, step=40, edge=w, square=False
    ):
        # check if a given tile intersects with
        # with the diagonal band of interest ...
        diag_from = tilej[0] - tilei[1]
        diag_to = tilej[1] - tilei[0]
        #
        band_from = 0
        band_to = band_idx
        # we are using this >2w trick to exclude
        # tiles from the lower triangle from calculations ...
        if (min(band_to, diag_to) - max(band_from, diag_from)) > 2 * w:
            # define origin:
            origin = (tilei[0], tilej[0])
            # RAW observed matrix slice:
            observed = mock_M_raw[slice(*tilei), slice(*tilej)]
            # trying new expected function:
            expected = mock_exp[slice(*tilei), slice(*tilej)]
            # for diagonal chuynking/tiling tilei==tilej:
            ice_weight_i = mock_v_ice[slice(*tilei)]
            ice_weight_j = mock_v_ice[slice(*tilej)]
            # that's the main working function from dotfinder:
            res = dotfinder.get_adjusted_expected_tile_some_nans(
                origin=origin,
                observed=observed,
                expected=expected,
                bal_weights=(ice_weight_i, ice_weight_j),
                kernels={"donut": kernel, "footprint": np.ones_like(kernel)},
                # nan_threshold=1,
                verbose=False,
            )
            is_inside_band = res["bin1_id"] > (res["bin2_id"] - band_idx)
            # new style, selecting good guys:
            does_comply_nans = res["la_exp." + "footprint" + ".nnans"] < nnans
            # so, select inside band and nNaNs compliant results and append:
            res_df = res_df.append(
                res[is_inside_band & does_comply_nans], ignore_index=True
            )

    # sort and reset index, there shouldn't be any duplicates now:
    res_df = (
        res_df
        .drop_duplicates()
        .sort_values(by=["bin1_id", "bin2_id"])
        .reset_index(drop=True)
    )

    # prepare mock_data for comparison:
    # apparently sorting is needed in this case:
    mock_res_sorted = (
        mock_res
        .drop_duplicates()
        .sort_values(by=["bin1_id", "bin2_id"])
        .reset_index(drop=True)
    )

    # ACTUAL TESTS:
    # integer part of DataFrame must equals exactly:
    assert res_df[["bin1_id", "bin2_id"]].equals(
        mock_res_sorted[["bin1_id", "bin2_id"]]
    )
    # compare floating point part separately:
    assert np.isclose(
        res_df["la_exp." + "donut" + ".value"],
        mock_res_sorted["la_expected"],
        equal_nan=True,
    ).all()



import pandas as pd
import numpy as np
from functools import partial
from .. import expected

import click
from .util import validate_csv
from . import cli

@cli.command()
@click.argument(
    "expected_path",
    metavar="EXPECTED_PATH",
    type=str,
    callback=partial(validate_csv, default_column="balanced.sum"),
)
@click.argument(
    "output_prefix",
    metavar="OUTPUT_PREFIX",
    type=str,
    nargs=1
)
@click.option(
    "--bins-per-order-magnitude",
    metavar="bins_per_order_magnitude",
    help="How many bins per order of magnitude. "
         "Default of 10 has a ratio of neighboring bins of about 1.25",
    type=int,
    nargs=1,
    default=10,
    show_default=True
)
@click.option(
    "--bin-layout",
    metavar="bin_layout",
    help="'fixed' means that bins are exactly the same for different datasets, "
          "and only depend on bins_per_order_magnitude "
         "'longest_regio' means that the last bin will end at size of the longest region. "
         "\nGOOD: the last bin will have as much data as possible. "
         "\nBAD: bin edges will end up different for different datasets, "
         "you can't divide them by each other",
    type=click.Choice(["fixed", "longest_region"]),
    nargs=1,
    default='fixed',
    show_default=True
)
@click.option(
    "--min-nvalid",
    metavar="min_nvalid",
    help="For each region, throw out bins (log-spaced) that have less than min_nvalid "
         "valid pixels. This will ensure that each entree in Pc by region has at least "
         "n_valid valid pixels. "
         "Don't set it to zero, or it will introduce bugs. Setting it to 1 is OK, but "
         "not recommended.",
    type=int,
    nargs=1,
    default=200,
    show_default=True
)
@click.option(
    "--min-count",
    metavar="min_count",
    help="If counts are found in the data, then for each region, throw out bins "
         "(log-spaced) that have more than min_counts of counts.sum (raw Hi-C counts). "
         "This will ensure that each entree in P(s) by region has at least min_count "
         "raw Hi-C reads",
    type=int,
    nargs=1,
    default=50,
    show_default=True
)
@click.option(
    "--spread-funcs",
    metavar="spread_funcs",
    help="A way to estimate the spread of the P(s) curves between regions. "
    "* 'minmax' - the minimum/maximum of by-region P(s)\n"
    "* 'std' - weighted standard deviation of P(s) curves (may produce negative results)\n "
    "* 'logstd' (recommended) weighted standard deviation in logspace",
    type=click.Choice(["minmax", "std", "logstd"]),
    default='logstd',
    show_default=True,
    nargs=1,
)
@click.option(
    "--spread-funcs-slope",
    metavar="spread_funcs_slope",
    help="Same as spread-funcs, but for slope (derivative) ratehr than P(s)",
    type=click.Choice(["minmax", "std", "logstd"]),
    default='std',
    show_default=True,
    nargs=1,
)
@click.option(
    "--resolution",
    metavar="resolution",
    help="Data resolution in bp. If provided, additonal column of separation in bp "
    "(s_bp) will be added to the outputs",
    type=int,
    nargs=1,
)
def logbin_expected(
    expected_path,
    output_prefix,
    bins_per_order_magnitude,
    bin_layout,
    min_nvalid,
    min_count,
    spread_funcs,
    spread_funcs_slope,
    resolution
):
    """
    Logarithmically bin expected values generated using compute_expected for cis data.
    
    This smoothes the data, resulting in clearer plots and more robust analysis results.
    Also calculates derivative after gaussian smoothing.
    For a very detailed escription, see
    https://github.com/open2c/cooltools/blob/51b95c3bed8d00a5f1f91370fc5192d9a7face7c/cooltools/expected.py#L988

    EXPECTED_PATH : The paths to a .tsv file with output of compute_expected.
    Must include a header. Use the '::' syntax to specify a summary column name.
    
    OUTPUT_PREFIX: Output file name prefix to store the logbinned expected
    (prefix.log.tsv) and derivative (prefix.der.tsv) in the tsv format."
    """

    # unpack expected path and name as generated by click's callback to validate_csv:
    expected_path, exp_summary_name = expected_path
    # that's what we expect as column names:
    expected_columns = ["region", "diag", "n_valid", "count.sum", exp_summary_name]
    if exp_summary_name == "count.sum":
        expected_columns = ["region", "diag", "n_valid", exp_summary_name]
    # expected dtype as a rudimentary form of validation:
    expected_dtype = {
        "region": np.str,
        "diag": np.int64,
        "n_valid": np.int64,
        "count.sum": np.float64,
        exp_summary_name: np.float64,
    }

    # use 'usecols' as a rudimentary form of validation,
    # and dtype. Keep 'comment' and 'verbose' - explicit,
    # as we may use them later:
    cvd = pd.read_csv(
        expected_path,
        usecols=expected_columns,
        dtype=expected_dtype,
        comment=None,
        sep='\t',
        verbose=False,
    )

    # name of the column with Probability of contacts is
    # based on the name of the column  with the diagonal-summary
    # stats in the input expected DataFrame:
    exp_summary_base, *_ = exp_summary_name.split(".")
    Pc_name = f"{exp_summary_base}.avg"

    lb_cvd, lb_slopes, lb_distbins = expected.logbin_expected(
        cvd,
        summary_name=exp_summary_name,
        bins_per_order_magnitude=bins_per_order_magnitude,
        bin_layout=bin_layout,
        min_nvalid=min_nvalid,
        min_count=min_count
    )
    # combine Probabilities of contact for the regions:
    lb_cvd_agg, lb_slopes_agg = expected.combine_binned_expected(
        lb_cvd,
        Pc_name=Pc_name,
        binned_exp_slope=lb_slopes,
        spread_funcs=spread_funcs,
        spread_funcs_slope=spread_funcs_slope
    )
    if resolution is not None:
        lb_cvd_agg['s_bp'] = lb_cvd_agg['diag.avg'] * resolution
        lb_slopes_agg['s_bp'] = lb_slopes_agg['diag.avg'] * resolution

    lb_cvd_agg.to_csv(
        f'{output_prefix}.log.tsv',
        sep="\t",
        index=False,
        na_rep="nan",
    )
    lb_cvd_agg.to_csv(
        f'{output_prefix}.der.tsv',
        sep="\t",
        index=False,
        na_rep="nan",
    )



import os.path as op

# import pandas as pd

import numpy as np
import pandas as pd
from numpy import testing

import bioframe
import cooler
from click.testing import CliRunner
import cooltools.expected
from cooltools.cli import cli

from itertools import combinations


# rudimentary expected functions for dense matrices:


def _diagsum_dense(matrix, ignore_diags=2, bad_bins=None):
    """
    function returning a diagsum list for a square symmetric and dense matrix
    it starts from the main diagonal and goes until the upper right element -
    the furthermost diagonal.
    """
    mat = matrix.copy().astype(float)
    if bad_bins is not None:
        mat[bad_bins, :] = np.nan
        mat[:, bad_bins] = np.nan

    # diagonal, starting from main one
    # all the way to the upper right element
    diags = range(len(mat))

    return [
        np.nanmean(mat.diagonal(i)) if (i >= ignore_diags) else np.nan for i in diags
    ]


def _diagsum_asymm_dense(matrix, bad_bin_rows=None, bad_bin_cols=None):
    """
    function returning a diagsum list for an arbitrary dense matrix, with no
    assumptions on symmetry.

    it starts from the bottom left element, goes through the "main" diagonal
    and to the upper right element - the frther most diagonal.
    """
    mat = matrix.copy().astype(float)
    if bad_bin_rows is not None:
        mat[bad_bin_rows, :] = np.nan
    elif bad_bin_cols is not None:
        mat[:, bad_bin_cols] = np.nan

    (mrows, mcols) = mat.shape

    # diagonals of an arbitrary rectangle matrix
    # negative diagonals are below the "main" one
    diags = range(-mrows + 1, mcols)

    return [np.nanmean(mat.diagonal(i)) for i in diags]
    # flatexp_tomat = lambda e,m_like: toeplitz(e[m_like.shape[0]-1::-1],e[m_like.shape[0]-1:])


def _blocksum_asymm_dense(matrix, bad_bin_rows=None, bad_bin_cols=None):
    """
    function returning a blocksum-based average for an arbitrary dense matrix,
    with no assumptions on symmetry.
    """
    mat = matrix.copy().astype(float)
    if bad_bin_rows is not None:
        mat[bad_bin_rows, :] = np.nan
    elif bad_bin_cols is not None:
        mat[:, bad_bin_cols] = np.nan

    return np.nanmean(mat)


# common parameters:
ignore_diags = 2
weight_name = "weight"
bad_bins = None
chunksize = 10_000  # keep it small to engage chunking
weight1 = weight_name + "1"
weight2 = weight_name + "2"
transforms = {"balanced": lambda p: p["count"] * p[weight1] * p[weight2]}


chromsizes = bioframe.fetch_chromsizes("mm9")
chromosomes = list(chromsizes.index)
supports = [(chrom, 0, chromsizes[chrom]) for chrom in chromosomes]

# test the most frequent use cases, balancing applied, no bad bins, etc.

common_regions = []
for i in range(4):
    chrom = chromosomes[i]
    halfway_chrom = int(chromsizes[chrom] / 2)
    reg1 = (chrom, 0, halfway_chrom)
    reg2 = (chrom, halfway_chrom, chromsizes[chrom])
    common_regions.append(reg1)
    common_regions.append(reg2)


def test_diagsum(request):
    # perform test:
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    res = cooltools.expected.diagsum(
        clr,
        regions=common_regions,
        transforms=transforms,
        weight_name=weight_name,
        bad_bins=bad_bins,
        ignore_diags=ignore_diags,
        chunksize=chunksize,
    )
    # calculate average:
    res["balanced.avg"] = res["balanced.sum"] / res["n_valid"]
    # check results for every "region"
    grouped = res.groupby("region")
    for name, group in grouped:
        matrix = clr.matrix(balance=weight_name).fetch(name)
        testing.assert_allclose(
            actual=group["balanced.avg"].values,
            desired=_diagsum_dense(matrix, ignore_diags=2),
            # rtol=1e-07,
            # atol=0,
            equal_nan=True,
        )


def test_diagsum_asymm(request):
    # perform test:
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    cis_pairwise_regions = [
        (r1, r2) for (r1, r2) in combinations(common_regions, 2) if (r1[0] == r2[0])
    ]
    regions1, regions2 = list(zip(*cis_pairwise_regions))
    res = cooltools.expected.diagsum_asymm(
        clr,
        regions1=regions1,
        regions2=regions2,
        transforms=transforms,
        weight_name=weight_name,
        bad_bins=bad_bins,
        chunksize=chunksize,
    )
    # calculate average:
    res["balanced.avg"] = res["balanced.sum"] / res["n_valid"]
    # check results for every block, defined as region1/2
    grouped = res.groupby(["region1", "region2"])
    for (name1, name2), group in grouped:
        matrix = clr.matrix(balance=weight_name).fetch(name1, name2)
        testing.assert_allclose(
            actual=group["balanced.avg"].values,
            desired=_diagsum_asymm_dense(matrix),
            # rtol=1e-07,
            # atol=0,
            equal_nan=True,
        )


def test_blocksum(request):
    # perform test:
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    all_pairwise_regions = combinations(common_regions, 2)
    regions1, regions2 = list(zip(*all_pairwise_regions))
    res = cooltools.expected.blocksum_asymm(
        clr,
        regions1=regions1,
        regions2=regions2,
        transforms=transforms,
        weight_name=weight_name,
        bad_bins=bad_bins,
        chunksize=chunksize,
    )
    # calculate average:
    res["balanced.avg"] = res["balanced.sum"] / res["n_valid"]
    # check results for every block , defined as region1/2
    # should be simplified as there is 1 number per block only
    grouped = res.groupby(["region1", "region2"])
    for (name1, name2), group in grouped:
        matrix = clr.matrix(balance=weight_name).fetch(name1, name2)
        testing.assert_allclose(
            actual=group["balanced.avg"].values,
            desired=_blocksum_asymm_dense(matrix),
            # rtol=1e-07,
            # atol=0,
            equal_nan=True,
        )


def test_expected_cli(request, tmpdir):
    # CLI compute-expected for chrom-wide cis-data
    in_cool = op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool")
    out_cis_expected = op.join(tmpdir, "cis.exp.tsv")
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "compute-expected",
            "--weight-name",
            weight_name,
            "-o",
            out_cis_expected,
            in_cool,
        ],
    )
    assert result.exit_code == 0
    clr = cooler.Cooler(in_cool)
    cis_expected = pd.read_table(out_cis_expected, sep="\t")
    grouped = cis_expected.groupby("region")
    # full chromosomes in this example:
    for chrom, group in grouped:
        matrix = clr.matrix(balance=weight_name).fetch(chrom)
        testing.assert_allclose(
            actual=group["balanced.avg"].values,
            desired=_diagsum_dense(matrix, ignore_diags=2),
            # rtol=1e-07,
            # atol=0,
            equal_nan=True,
        )


def test_expected_regions_cli(request, tmpdir):
    # CLI compute expected for cis-data with arbitrary regions
    # which may overlap. But it is symmetrical cis-case.
    in_cool = op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool")
    in_regions = op.join(request.fspath.dirname, "data/mm9.named_overlap_regions.bed")
    out_cis_expected = op.join(tmpdir, "cis.regions.exp.tsv")
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "compute-expected",
            "--weight-name",
            weight_name,
            "--regions",
            in_regions,
            "-o",
            out_cis_expected,
            in_cool,
        ],
    )
    assert result.exit_code == 0
    clr = cooler.Cooler(in_cool)
    cis_expected = pd.read_csv(out_cis_expected, sep="\t")
    grouped = cis_expected.groupby("region")
    # deal with named and overlapping regions here:
    regions_df = pd.read_csv(in_regions, sep="\t", header=None)
    regions_df = regions_df.set_index(3)
    for region_name, group in grouped:
        ucsc_region = regions_df.loc[region_name].to_list()
        matrix = clr.matrix(balance=weight_name).fetch(ucsc_region)
        testing.assert_allclose(
            actual=group["balanced.avg"].values,
            desired=_diagsum_dense(matrix, ignore_diags=2),
            # rtol=1e-07,
            # atol=0,
            equal_nan=True,
        )


def test_trans_expected_regions_cli(request, tmpdir):
    # CLI compute expected for cis-data with arbitrary regions
    # which may overlap. But it is symmetrical cis-case.
    in_cool = op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool")
    in_regions = op.join(
        request.fspath.dirname, "data/mm9.named_nonoverlap_regions.bed"
    )
    out_trans_expected = op.join(tmpdir, "cis.regions.exp.tsv")
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "compute-expected",
            "--weight-name",
            weight_name,
            "--regions",
            in_regions,
            "--contact-type",
            "trans",
            "-o",
            out_trans_expected,
            in_cool,
        ],
    )
    assert result.exit_code == 0
    clr = cooler.Cooler(in_cool)
    trans_expected = pd.read_csv(out_trans_expected, sep="\t")
    # grouped = trans_expected.groupby("region1","region2")
    trans_expected = trans_expected.set_index(["region1", "region2"])
    # deal with named and overlapping regions here:
    regions_df = pd.read_csv(in_regions, sep="\t", header=None)
    # prepare pairwise combinations of regions for trans-expected (blocksum):
    regions_pairwise = combinations(regions_df.itertuples(index=False), 2)
    regions1, regions2 = zip(*regions_pairwise)
    regions1 = pd.DataFrame(regions1)
    regions2 = pd.DataFrame(regions2)
    for i in range(len(trans_expected)):
        region1_name = regions1.iloc[i, 3]
        region2_name = regions2.iloc[i, 3]
        ucsc_region1 = regions1.iloc[i, :3].to_list()
        ucsc_region2 = regions2.iloc[i, :3].to_list()
        matrix = clr.matrix(balance=weight_name).fetch(ucsc_region1, ucsc_region2)
        testing.assert_allclose(
            actual=trans_expected.loc[(region1_name, region2_name), "balanced.avg"],
            desired=_blocksum_asymm_dense(matrix),
            # rtol=1e-07,
            # atol=0,
            equal_nan=True,
        )


def test_logbin_expected_cli(request, tmpdir):
    # test CLI logbin-expected for default chrom-wide output of compute-expected
    in_cool = op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool")
    out_cis_expected = op.join(tmpdir, "cis.exp.tsv")
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "compute-expected",
            "--weight-name",
            weight_name,
            "-o",
            out_cis_expected,
            in_cool,
        ],
    )
    assert result.exit_code == 0

    # consider adding logbin expected baswed on raw counts
    binsize = 1_000_000
    logbin_prefix = op.join(tmpdir, "logbin_prefix")
    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["logbin-expected", "--resolution", binsize, out_cis_expected, logbin_prefix],
    )
    assert result.exit_code == 0
    # make sure logbin output is generated:
    assert op.isfile(f"{logbin_prefix}.log.tsv")
    assert op.isfile(f"{logbin_prefix}.der.tsv")


def test_logbin_interpolate_roundtrip():
    from cooltools.expected import (
        logbin_expected,
        combine_binned_expected,
        interpolate_expected,
    )

    N = [100, 300]
    diag = np.concatenate([np.arange(i, dtype=int) for i in N])
    region = np.concatenate([i * [j] for i, j in zip(N, ["chr1", "chr2"])])
    prob = 1 / (diag + 1)
    n_valid = prob * 0 + 300
    count_sum = n_valid * prob * 100
    balanced_sum = n_valid * prob
    df = pd.DataFrame(
        {
            "region": region,
            "diag": diag,
            "n_valid": n_valid,
            "count.sum": count_sum,
            "balanced.sum": balanced_sum,
        }
    )
    df2 = df.copy()
    df2["balanced.sum"].values[: N[0]] *= 2  # chr2 is different  in df2

    # simple roundtrip
    exp1, der1, bins = logbin_expected(df2)
    interp = interpolate_expected(df2, exp1)
    max_dev = np.nanmax(
        np.abs(
            (interp["balanced.avg"] - df2["balanced.sum"] / df2["n_valid"])
            / interp["balanced.avg"]
        )
    )
    assert max_dev < 0.1  # maximum absolute deviation less than 10%

    # using combined expected succeeds when chroms are same
    exp1, der1, bins = logbin_expected(df)
    comb, cder = combine_binned_expected(exp1, der1)
    interp = interpolate_expected(df, comb, by_region=False)
    max_dev = np.nanmax(
        np.abs(
            (interp["balanced.avg"] - df["balanced.sum"] / df["n_valid"])
            / interp["balanced.avg"]
        )
    )
    assert max_dev < 0.1  # maximum absolute deviation less than 10%

    # using combined expected deemed to fail with df2
    exp1, der1, bins = logbin_expected(df2)
    comb, cder = combine_binned_expected(exp1, der1)
    interp = interpolate_expected(df2, comb, by_region=False)
    max_dev = np.nanmax(
        np.abs(
            (interp["balanced.avg"] - df2["balanced.sum"] / df2["n_valid"])
            / interp["balanced.avg"]
        )
    )
    assert max_dev > 0.1  # we are now different because we fed combined expected

    # using chr2 for chr1 works if they are the same
    interp = interpolate_expected(df, exp1, by_region="chr2")
    max_dev = np.abs(
        (interp["balanced.avg"] - df["balanced.sum"] / df["n_valid"])
        / interp["balanced.avg"]
    )
    assert np.nanmax(max_dev) < 0.1  # maximum absolute deviation less than 10%

    # using chr2 for chr1 fails if chroms are different
    interp = interpolate_expected(df2, exp1, by_region="chr2")
    max_dev = np.abs(
        (interp["balanced.avg"] - df2["balanced.sum"] / df2["n_valid"])
        / interp["balanced.avg"]
    )
    assert (
        np.nanmax(max_dev[N[0] :]) < 0.1
    )  # we are now correct for chr2 because we chose chr2
    assert np.nanmax(max_dev[: N[0]]) > 0.5  # we are now different for chr1


def test_diagsum_from_array():
    from cooltools.expected import diagsum_from_array

    # Toy data: create a uniform 1/s decay as input
    ar = np.arange(100)
    ar = np.abs(ar[:, None] - ar[None, :])  # like toeplitz(ar)
    ar[ar == 0] = 1
    ar = 1 / ar

    # Simple symmetric case
    exp = _diagsum_dense(ar, ignore_diags=0, bad_bins=None)
    exp1 = diagsum_from_array(ar, ignore_diags=0)
    exp1["balanced.avg"] = exp1["balanced.sum"] / exp1["n_valid"]
    assert np.allclose(exp, exp1["balanced.avg"].values, equal_nan=True)

    # Assymetric region
    exp = _diagsum_asymm_dense(ar[:50, 50:], bad_bin_rows=None, bad_bin_cols=None)
    exp1 = diagsum_from_array(ar[:50, 50:], ignore_diags=0, offset=(0, 50))
    exp1["balanced.avg"] = exp1["balanced.sum"] / exp1["n_valid"]
    assert np.allclose(exp, exp1["balanced.avg"].values, equal_nan=True)

    # Add some "bad" bins. Should have same outcome for this synthetic dataset
    # because input was homogenous decay.
    ar[3:5] = 0
    ar[:, 3:5] = 0
    exp = _diagsum_dense(ar, ignore_diags=0, bad_bins=list(range(3, 5)))
    exp1 = diagsum_from_array(ar, ignore_diags=0)
    exp1["balanced.avg"] = exp1["balanced.sum"] / exp1["n_valid"]
    assert np.allclose(exp, exp1["balanced.avg"].values, equal_nan=True)



import warnings
import os
import numpy as np
import joblib
from . import eigdecomp
import cooler


# ==============================================================
#  CONTRAST FUNCTIONS
# ==============================================================


def contrast_diags(
    M,
    modality="AnyAny_vs_Mixed",
    modality_params=None,
    I=None,
    ignore_diags=1,
    exclude_nans_from_paircounts=True,
    i0=None,
    i1=None,
    phasing_track=None,
    normalize=False,
    verbose=False,
):
    """
    Compute the contrast in M between different sets of pixels specified by
    modality, modality_params or by I, if given.

    Returns the contrast in diagonals with offset s=0..len(M), and the
    weighted average over all s. If 'normalize' is True, contrast is
    normalized to be in [-1,1].

    Parameters
    ----------
    M: 2d array
        Typically a contact matrix.

    modality: string, (optional, default: "AnyAny_vs_Mixed")
        Specifies the contrast modality, potentially together ``modality_params``.
        See :func:`constrast.indicatormat` for valid modalities.

    modality_params: list, (optional if modality=='AnyAny_vs_mixed' or None)
        Parameters required by 'modality'. See :func:`contrast.indicatormat()`
        for valid modalities and required parameters.

        Typical format: ``[bin_types, Type]``, where
            bin_types: 1d array
                types of the bins, e.g. [0,0,1,1,1,0,...]
                Note: the numerical values identifying the types are
                inconsequential they are only used to classify the loci into
                groups.
            Type: number
                one of the values in bin_types: singles out a specific type

        If None, modality_params = [EV>np.nanmean(EV)],
            with EV computed here: eigdecomp.cis_eig(M, phasing_track=phasing_track)[1][0]

    I: indicator matrix, optional
        Specifies the two sets of pixels from which to compute contrast in M.
        If given, this overrides ``modality`` and ``modality_params``.
        If None, computed here as ``contrast.indicatormat(modality, modality_params)``.

    normalize: boolean, optional
        If True, contr_diags is normalized as follows
            contr_norm[s]
                = (contr[s]-1)/(contr[s]+1)
                = (set1_avintens[s] - set2_avintens[s]) / (set1_avintens[s] + set2_avintens[s]).
        The weighted mean, contr, is then computed from contr_norm[s].

    ignore_diags: integer, optional
        Number of diagonals in M to be ignored.

    exclude_nans_from_paircounts: boolean, optional
        If True, pixels with NaN in M are not counted towards valid pixels.

    i0, i1: integers, optional
        Will use (trimmed by len(M)):
            M[i0:i1,i0:i1]
            v[i0:i1],
            phasing_track[i0:i1]
        Note: EV is computed from restriced M if modality_params is None

    phasing_track: 1D numpy array, optional
        len(phasing_track) must be len(M)
        to flip EVs
        only used if modality_params is None

    Returns
    -------
    contr_diags: 1D numpy array, length M
        Contrast in diagonals with offsets s=0..len(M)

    contr: float
        Weighted mean of contrast_diags.
        np.nansum(contrast*p)/np.nansum(p)
        with weights (#set1_pixels(s)*#set2_pixels(s))**.5

    I: 2D array
        The computed (or supplied) indicator matrix.

    modality_params: list
        The computed (or supplied) modality_params useful if modality_params
        was computed here.

    additional_info: list
        [
            set1_intens: 1D array,
                nansum(diag(I*M,s)), i.e. total intensity in set1 for all s
            set1_pixels:  1D array,
                nansum(diag(I)) i.e. number of valid pixels in set1 for all s
            set2_intens: 1D array,
                nansum(diag((1-I)*M)), i.e. total intensity in set2 for all s
            set2_pixels:  1D array,
                nansum(1-diag(I)) i.e. number of valid pixels in set2 for all s
        ]

    """
    # get matrix
    if isinstance(M, np.ndarray):
        if len(M.shape) != 2:
            raise ValueError("M was an array, but dimensionality was not 2")
    else:
        try:
            if verbose:
                print("... loading M ...")
            M = joblib.load(M)
        except:
            raise ValueError("cmap could not be loaded")

    # trim M,I and phasing track
    M = M[i0:i1, i0:i1]
    if I is not None:
        I = I[i0:i1, i0:i1]
    if phasing_track is not None:
        phasing_track = phasing_track[i0:i1]

    # get indicator matirx (i.e. the two sets of pixels to be compared)
    if I is None:
        if modality_params is None:
            if modality == "AnyAny_vs_Mixed":
                if verbose:
                    print("... getting types from EV...")
                bin_types = np.ones(len(M)) * np.nan
                EV = eigdecomp.cis_eig(M, phasing_track=phasing_track)[1][0]
                val_inds = np.isfinite(EV)
                get_bin_identities = lambda x: x > np.nanmean(x)
                bin_types[val_inds] = get_bin_identities(EV[val_inds])
                modality_params = [bin_types]
            else:
                raise ValueError(
                    "modality_params=None only allowed with modality='AnyAny_vs_Mixed'. "
                    "For all other modalityes modality_params has to be given (a list). "
                    "See contrast.indicatormatrix() for valid modalities and "
                    "required modality_params."
                )

        if verbose:
            print("... constructing indicator matrix ...")
        I = indicatormat(modality, modality_params)

    # get contrast
    if verbose:
        print("... computing contrast ...")
    contr_diags, add_info = contrast_diags_indicatormatrix(
        M,
        I,
        ignore_diags=ignore_diags,
        exclude_nans_from_paircounts=exclude_nans_from_paircounts,
        normalize=normalize,
    )

    # weighted average over diagonals
    p = (
        add_info[1] * add_info[3]
    ) ** 0.5  # weights of diagonals: (#set1_pairs(s)*#set2_pairs(s))**.5
    contr = np.nansum(contr_diags * p) / np.nansum(p)

    return contr_diags, contr, I, modality_params, add_info


def contrast_diags_indicatormatrix(
    M,
    I,
    ignore_diags=0,
    exclude_nans_from_paircounts=True,
    normalize=True,
    verbose=False,
):
    """
    Computes, for all upper diagonals, the 'contrast' in M, namely the ratio
    of average intensities (values) in M in two sets of pixels.

    More formally, for diagonals with offsets s=0..len(M):
        contr[s] = <diag(M,s)>_set1 / <diag(M,s)>_set2)

    If normalize is True, the contrast is normalized as follows:
        contr_norm[s]
            = (conts[s]-1)/(contr[s]+1)
            = (<diag(M,s)>_set1 - <diag(M,s)>_set2) /
              (<diag(M,s)>_set1 + <diag(M,s)>_set2)

    The normalized contrast is between -1 and 1.

    The two sets of pixels are specified by the indicator matrix I as follows:
        set1:    all pixels (i,j) with I[i,j]=1
        set2:    all pixels (i,j) with I[i,j]=0
        neither: all pixels (i,j) with I[i,j]=nan

    If exclude_nans_from_paircounts==True, pixels M(i,j)=nan entail I(i,j)=nan
    (such that those pixels don't count towards neither set1 nor set2).

    Parameters
    ----------
    M: 2D numpy array
        the data, eg. a Hi-C contact matrix

    I: 2D numpy array
        indicator matrix specifying the two sets of pixels
        can contain only
        - zeros (for set1),
        - ones (for set2),
        - NANs (neither)
        otherwise a value error is raised
        must have same shape as M

    ignore_diags: int
        the entries [0:ignore_diags] will be NANs in all returns

    exclude_nans_from_paircounts: boolean
        if True, I(i,j) gets nans for all pixels (i,j) where M is nan

    normalize: boolean, optional
        if True, contrast is normalized as follows:
        contr_norm[s] = (contr[s]-1)/(contr[s]+1)
                      = (set1_avintens[s] - set2_avintens[s]) /
                        (set1_avintens[s] + set2_avintens[s])

    Returns
    -------
    contr_diags: 1D array
        ratio of mean values in M of set1 and set2, by diagonal s

    additional_info: list
        [
            set1_intens: 1D array,
                nansum(diag(I*M,s)), i.e. total intensity in set1 for all s
            set1_pixels:  1D array,
                nansum(diag(I)) i.e. number of valid pixels in set1 for all s
            set2_intens: 1D array,
                nansum(diag((1-I)*M)), i.e. total intensity in set2 for all s
            set2_pixels:  1D array,
                nansum(1-diag(I)) i.e. number of valid pixels in set2 for all s
        ]

    Notes
    -----
    * Lower halves of M and I are ignored but a warning is thrown if M is not
      symmetric.
    * The bare contrast can be computed from the normalized one as
      contr = (contr_norm + 1)/(contr_norm - 1).

    Examples
    --------
    M = [3 2 1        I = [1 1 0
         . 1 1             . 1 0
         . . 1]            . . nan]

    The resulting below quantities are by diagonal with increasing offset
    s=0..len(M):

    set1_intens    = [4, 2, 0]
    set1_pixels    = [2, 1, 0]
    =>
    set1_avintens  = [2, 2, nan]

    set2_intens    = [0, 1, 1]
    set1_pixels    = [0, 1, 1]
    =>
    set2_avintens  = [nan, 1, 1]
    =>
    contrast_diags_indicatormatrix(M, I, ignore_diats=0)
        = set1_avintens / set2_avintens
        = [nan, 2, nan]
          (indeed, only 1st off-diagonal has both types, others have to be nans)

    contrast_diags_indicatormatrix(M, I, ignore_diats=0, normalize=True)
        = (set1_avintens - set2_avintens) / (set1_avintens + set2_avintens)
        = [nan, 1/3, nan]
          indeed, only 1st off-diagonal has both types, others have to be nans

    """

    # check if M is symmetric
    tol = 1e-8
    if not np.allclose(M, M.T, atol=tol, equal_nan=True):
        warnings.warn(
            "M is not symmetric to within {}, " "I'm using the upper half".format(tol)
        )

    # check if I has only 0,1,nan
    Iu = np.unique(I)
    nanIu = Iu[~np.isnan(Iu)]
    illegalvals = [i for i in nanIu if i not in [0.0, 1.0]]
    if np.any(illegalvals):
        raise ValueError(
            "I is not an indicator matrix: "
            "it contains value(s) other than {0,1,np.nan}:",
            illegalvals,
        )

    # check if M and I have same shape
    if not M.shape == I.shape:
        raise ValueError("M and I must have the same shape")

    L = len(M)

    # make a copy of indicator matrix
    Iuse = np.copy(I).astype(float)

    # exclude pixels with M=np.nan from counts towards valid pairs
    if exclude_nans_from_paircounts:
        Iuse[np.isnan(M)] = np.nan

    # compute valid paris, counts and contrast
    set1_pixels = np.zeros(L) * np.nan
    set1_intens = np.zeros(L) * np.nan
    set2_pixels = np.zeros(L) * np.nan
    set2_intens = np.zeros(L) * np.nan

    if verbose:
        print("... ... starting to loop over s")
        report_progres_points = np.linspace(0, L, 11).astype(int)
    for s in range(ignore_diags, L):
        if verbose and s in report_progres_points:
            print("... ... {}% done".format(int(s / L * 100)))
        diagI = np.diag(Iuse, s)  # get diagonal s of indicator matrix
        diagM = np.diag(M, s)  # get diagonal s of data matrix

        set1_pixels[s] = np.nansum(diagI)  # number of valid pixels in set 1
        set2_pixels[s] = np.nansum(1 - diagI)  # number of valid pixels in set 2

        set1_intens[s] = np.nansum(diagM * diagI)
        set2_intens[s] = np.nansum(diagM * (1 - diagI))

    with np.errstate(invalid="ignore"):
        # intensity average of valid pixels in set 1 in diagonal s
        set1_avintens = set1_intens / set1_pixels
        # intensity average of valid pixels in set 2 in diagonal s
        set2_avintens = set2_intens / set2_pixels

        if normalize is True:
            contr = (set1_avintens - set2_avintens) / (set1_avintens + set2_avintens)
        else:
            contr = set1_avintens / set2_avintens

    additional_info = [set1_intens, set1_pixels, set2_intens, set2_pixels]

    return contr, additional_info


# ==============================================================
#  MISCELLANEOUS FUNCTIONS
# ==============================================================


def diagcounts(I, vals=[1, 0]):
    """
    count occurrances of each v in vals in all diagonals of I with offsets s=0..len(I)

    Returns
    -------
    list of 1D arrays of floats, one for each v in vals, each of len=len(I)
        default: count 1s and 0s in each diagonal:
                 return [S1,S0] with:
                 S1: nansum(diag(I,s)==1) for each offset s=0..len(I)
                 S0: nansum(diag(I,s)==0) for each offset s=0..len(I)
    """
    L = len(I)
    diagcounts = []
    for i in range(len(vals)):
        S = np.zeros(L)
        for s in range(L):
            S[s] = np.nansum(np.diag(I, s) == vals[i])
        diagcounts.append(S)
    return diagcounts


def get_types(v, vals=None):
    """
    Find unique and finite values (bin types) in v.
    If vals is not None, restrict to types in vals (floats and ints are
    considered identical).

    Parameters
    ----------
    v: sequence of numbers
        a vector of "types" used to construct indicator matrices

    vals: sequence of numbers, optional
        default: unique and finite values in v

    Returns
    -------
    valid_types: 1d array

    """
    v = np.asarray(v)
    types = np.unique(v[np.isfinite(v)])  # find types (cast to floats)

    if vals is not None:
        vals = np.asarray(vals)
        vals = np.unique(vals[np.isfinite(vals)])  # cast to floats
        valid_types = np.asarray([t for t in types if t in vals and np.isfinite(t)])
    else:
        valid_types = types

    return valid_types


def discretize_track(v, get_bin_identities=lambda x: x > np.nanmean(x)):
    """
    discretizes a (quasi)continuous track (eg. an EV)
    using the supplied lambda function
    by default: split by nanmean

    returns: discretized track, non-finite values are left intact
    """
    v_disc = np.ones(len(v)) * np.nan
    val_inds = np.isfinite(v)  # valid indices
    v_disc[val_inds] = get_bin_identities(v[val_inds])
    return v_disc


def normalize(v):
    # normalization
    v = np.asarray(v)
    return (v - 1) / (v + 1)


def normalize_inv(v):
    # invert normalization
    v = np.asarray(v)
    return (1 + v) / (1 - v)


# ==============================================================
#  GENERAL PURPOSE INDICATOR MATRIX FUNCTION
# ==============================================================


def indicatormat(modality=None, params=None):
    """
    get indicatormatrix I with a given modality by calling
    the function specified by 'modality'

    Parameters
    ----------
    modality: string, optional
        valid modalities are listed below in dict 'valid_modalities'
        if modality is None: print a list of valid modalities and return nothing
    params: list, optional
        have to match the parameters of the called indicator matrix functions

    Returns
    -------
    I: 2d array
        the computed indicator matrix

    """

    # NOTE: when adding/removing a specific indicator function below, also add/remove it here
    valid_modalities = {
        "AnyAny_vs_Mixed": indicatormat_AnyAny_vs_Mixed,  # params: [v]
        "TypeType_vs_Mixed": indicatormat_TypeType_vs_Mixed,  # params: [v, Type]
        "TypeType_vs_TypeOther": indicatormat_TypeType_vs_TypeOther,  # params: [v, Type]
        "TypeType_vs_NontypeNontype": indicatormat_TypeType_vs_NontypeNontype,  # params: [v, Type]
        "TypeType_vs_Rest": indicatormat_TypeType_vs_Rest,  # params: [v, Type]
        "Segments_vs_Rest": indicatormat_Segments_vs_Rest,  # params: [segments, L=None, bad_bins=[]]
    }

    valid_modalities_names = ""
    for k in valid_modalities.keys():
        valid_modalities_names = valid_modalities_names + k + "\n"

    if modality is None:
        print("valid modalities are:\n=====================\n" + valid_modalities_names)
        return

    if modality not in valid_modalities.keys():
        raise ValueError(
            modality
            + " is not a valid modality. Valid modalities are: \n"
            + valid_modalities_names
        )

    I = valid_modalities[modality](*params)

    return I


# ==============================================================
#  INDICATOR MATRIX FUNCTIONS
#         FOR SPECIFIC COMPARTMENTALIZATION MODALITIES
# ==============================================================

# NOTE: when adding/removing a specific indicator function,
#       also add/remove it to 'valid_modalities' in 'indicatormat()'


def indicatormat_AnyAny_vs_Mixed(v):
    """
    Parameters
    ----------
    v: sequence of numbers
        a vector of "types" used to construct indicator matrices
        types are the unique and finite values in v

    Returns
    -------
    I: 2d array
        indicator matrix with the following properties
        I[i,j]=1    if v[i]==v[j],               i.e. thistype-to-thistype for any type
        I[i,j]=nan  if v[i]==nan or v[j]==nan,   i.e. excluding invalid bins
        I[i,j]=0    otherwise,                   i.e. thistype-to-othertype for valid types

    """
    v = np.asarray(v)
    L = v.size

    types = get_types(v)

    # put nans in rows and cols with type==nan, zero elsewhere
    v_aux = np.zeros(L) * np.nan
    v_aux[~np.isnan(v)] = 0
    I = np.outer(v_aux, v_aux)  # => I[i,j]=nan if v[i]==nan or v[j]==nan

    # put 1 in I[i,j] if type[v[i]]==type[v[j]]
    for thistype in types:
        v_aux = np.zeros(L) * np.nan
        v_aux[v == thistype] = 1
        I_thistype = np.outer(
            v_aux, v_aux
        )  # => I_thistype: 1 for (i,j) both of this type, nan elsewhere
        I[
            I_thistype == 1
        ] = 1  # => I_anytype: zero for across types, one for withing any type, nan elsewhere

    return I


def indicatormat_TypeType_vs_Mixed(v, Type):
    """
    Parameters
    ----------
    v: sequence of numbers
        a vector of "types" used to construct indicator matrices

    Type: number
        type for which to compute indicator matrix with properties

    Returns
    -------
    I: 2d array:
        indicator matrix with the following properties
        I[i,j]=1    if v[i]==v[j]==val,    i.e. thistype-to-thistype
        I[i,j]=0    if v[i]~=v[j]==val,    i.e. sometype-to-othertype
        I[i,j]=nan  otherwise              i.e. othertype-to-othertype and invalid bins

    """
    v = np.asarray(v)
    L = v.size

    if not np.any(v == Type):
        warnings.warn("Type not found in v")

    I_AnyAny_vs_Mixed = indicatormat_AnyAny_vs_Mixed(
        v
    )  # get auxiliury matrix AnyAny_vs_AnyOther

    v_aux = np.zeros(L) * np.nan
    v_aux[v == Type] = 1
    I = np.outer(v_aux, v_aux)  # 1 for (i,j) both of this type, nan elsewhere
    I[I_AnyAny_vs_Mixed == 0] = 0  # 0 for (i,j) of different types

    return I


def indicatormat_TypeType_vs_TypeOther(v, Type):
    """
    Parameters
    -----------
    v: sequence of numbers
        a vector of "types" used to construct indicator matrices

    Type: number
        type for which to compute indicator matrix with properties

    Returns
    -------
    I: 2d array:
        indicator matrix with the following properties
        I[i,j]=1    if v[i]==v[j]==val,         i.e. thistype-to-thistype
        I[i,j]=0    if v[i]==val xor v[j]==val  i.e. thistype-to-othertype
        I[i,j]=nan  otherwise                   i.e. othertype-to-othertype and invalid bins

    """
    v = np.asarray(v)
    L = v.size

    if not np.any(v == Type):
        warnings.warn("Type not found in v")

    if 0:
        # get auxiliury matrix AnyAny_vs_AnyOther
        I_AnyAny_vs_Mixed = indicatormat_AnyAny_vs_Mixed(v)[0]

        v_aux = np.zeros(L) * np.nan
        v_aux[v == Type] = 1
        I = np.outer(v_aux, v_aux)  # 1 for (i,j) both of this type, nan elsewhere
        for i, j in np.ndindex((L, L)):
            if (
                ((v[i] == Type) != (v[j] == Type))
                and np.isfinite(v[i])
                and np.isfinite(v[j])
            ):
                I[i, j] = 0

    if 1:
        # put nans in rows and cols with type==nan, zero elsewhere
        v_aux = np.zeros(L) * np.nan
        v_aux[~np.isnan(v)] = 0
        I_invalid = np.outer(v_aux, v_aux)  # => I[i,j]=nan if v[i]==nan or v[j]==nan

        # put zeros in stripes that contain Type
        v_aux = np.ones(L)
        v_aux[v == Type] = 0
        I_striped = np.outer(v_aux, v_aux)  # 0 if v[i]==Type or v[j]==Type, 1 elsewhere

        # put ones in rectangles that are both Type
        v_aux = np.zeros(L) * np.nan
        v_aux[v == Type] = 1
        I_TypeType = np.outer(v_aux, v_aux)  # 1 if v[i]==v[j]==Type, nan elsewhere

        I = np.zeros((L, L)) * np.nan
        I[I_striped == 0] = 0
        I[I_TypeType == 1] = 1
        I[np.isnan(I_invalid)] = np.nan

    return I


def indicatormat_TypeType_vs_NontypeNontype(v, Type):
    """
    Parameters
    -----------
    v: sequence of numbers
        a vector of "types" used to construct indicator matrices

    Type: number
        type for which to compute indicator matrix with properties

    Returns
    -------
    I: 2d array:
        indicator matrix with the following properties
        I[i,j]=1    if v[i]==v[j]==val,         i.e. thistype-to-thistype
        I[i,j]=0    if v[i]!=val and v[j]!=val  i.e. anything not involving type
        I[i,j]=nan  otherwise                   i.e. type-to-othertype and invalid bins

    """
    v = np.asarray(v)
    L = v.size

    if not np.any(v == Type):
        warnings.warn("Type not found in v")

    # get auxiliury matrix AnyAny_vs_AnyOther
    I_AnyAny_vs_Mixed = indicatormat_AnyAny_vs_Mixed(v)[0]

    v_aux = np.zeros(L) * np.nan
    v_aux[v == Type] = 1
    I = np.outer(v_aux, v_aux)  # 1 for (i,j) both of this type, nan elsewhere

    v_aux = np.zeros(L) * np.nan
    v_aux[(v != Type) * np.isfinite(v)] = 1
    I_aux = np.outer(v_aux, v_aux)
    I[I_aux == 1] = 0

    return I


def indicatormat_TypeType_vs_Rest(v, Type):
    """
    Parameters
    -----------
    v: sequence of numbers
        a vector of "types" used to construct indicator matrices

    Type: number
        type for which to compute indicator matrix with properties

    Returns
    -------
    I: 2d array
        indicator matrix with the following properties
        I[i,j]=1    if v[i]==v[j]==val,          i.e. thistype-to-thistype
        I[i,j]=nan  if v[i]==nan or v[j]==nan,   i.e. excluding invalid bins
        I[i,j]=0    otherwise,                   i.e. thistype-to-othertype and anytype-to-anyother

    """

    v = np.asarray(v)
    L = v.size

    if not np.any(v == Type):
        warnings.warn("Type not found in v")

    # get auxiliury matrix AnyAny_vs_AnyOther
    I_AnyAny_vs_Mixed = indicatormat_AnyAny_vs_Mixed(v)[0]

    # construc indicator matrices
    v_aux = np.zeros(L) * np.nan
    v_aux[np.isfinite(v)] = 0
    v_aux[v == Type] = 1
    I = np.outer(v_aux, v_aux)  # 1 for (i,j) both of this type, nan elsewhere

    return I


def indicatormat_Segments_vs_Rest(segments, L=None, bad_bins=[]):
    """
    Parameters
    ----------
    segments: list of tuples
        [(s0,e0),(s1,e1), ...]
        where (s,e) are the start and end points of segments
        (endpoints are exclusive)
        Note: endpoints are trimmed to L

    L: integer, optional
        default: last endpoint

    bad_bins: list of integers, optional
        bad bins, rows and comuns get np.nan throughout

    Returns
    -------
    I: 2d array
        indicator matrices with the following properties
        I[i,j]=1    if i and j within same segment
        I[i,j]=nan  if i in bad_bins or j in bad_bins
        I[i,j]=0    otherwise

    """
    if L is None:
        L = segments[-1][1]

    v_aux = np.zeros(L)
    if bad_bins != []:
        v_aux[np.asarray(bad_bins)] = np.nan
    I = np.outer(v_aux, v_aux)

    segments_trimmed = []
    for s, e in segments:
        e = min(e, L)
        segments_trimmed.append((s, e))
        I[s:e, s:e] = I[s:e, s:e] + 1
    I[I > 1] = 1

    return I



import numpy as np
import scipy
import scipy.stats

import pandas as pd
from .lib import numutils

import bioframe


def _phase_eigs(eigvals, eigvecs, phasing_track, sort_metric=None):
    """
    Flip `eigvecs` to achieve a positive correlation with `phasing_track`.

    Parameters
    ----------
    sort_metric : str
        If provided, re-sort `eigenvecs` and `eigvals` in the order of
        decreasing correlation between phasing_track and eigenvector, using the
        specified measure of correlation. Possible values:
        'pearsonr' - sort by decreasing Pearson correlation.
        'var_explained' - sort by decreasing absolute amount of variation in
        `eigvecs` explained by `phasing_track` (i.e. R^2 * var(eigvec)).
        'MAD_explained' - sort by decreasing absolute amount of Median Absolute
        Deviation from the median of `eigvecs` explained by `phasing_track`
        (i.e. COMED(eigvec, phasing_track) * MAD(eigvec)).
        'spearmanr' - sort by decreasing Spearman correlation.
    """

    corrs = []
    for eigvec in eigvecs:
        mask = np.isfinite(eigvec) & np.isfinite(phasing_track)
        if sort_metric is None or sort_metric == "spearmanr":
            corr = scipy.stats.spearmanr(phasing_track[mask], eigvec[mask])[0]
        elif sort_metric == "pearsonr":
            corr = scipy.stats.pearsonr(phasing_track[mask], eigvec[mask])[0]
        elif sort_metric == "var_explained":
            corr = scipy.stats.pearsonr(phasing_track[mask], eigvec[mask])[0]
            # multiply by the sign to keep the phasing information
            corr = np.sign(corr) * corr * corr * np.var(eigvec[mask])
        elif sort_metric == "MAD_explained":
            corr = numutils.COMED(phasing_track[mask], eigvec[mask]) * numutils.MAD(
                eigvec[mask]
            )
        else:
            raise ValueError("Unknown sorting metric: {}".format(sort_metric))

        corrs.append(corr)

    # flip eigvecs
    for i in range(len(eigvecs)):
        eigvecs[i] = np.sign(corrs[i]) * eigvecs[i]

    # sort eigvecs
    if sort_metric is not None:
        idx = np.argsort(-np.abs(corrs))
        eigvals, eigvecs = eigvals[idx], eigvecs[idx]

    return eigvals, eigvecs


def cis_eig(
    A, n_eigs=3, phasing_track=None, ignore_diags=2, clip_percentile=0, sort_metric=None
):
    """
    Compute compartment eigenvector on a dense cis matrix.

    Note that the amplitude of compartment eigenvectors is weighted by their
    corresponding eigenvalue

    Parameters
    ----------
    A : 2D array
        balanced dense contact matrix
    n_eigs : int
        number of eigenvectors to compute
    phasing_track : 1D array, optional
        if provided, eigenvectors are flipped to achieve a positive correlation
        with `phasing_track`.
    ignore_diags : int
        the number of diagonals to ignore
    clip_percentile : float
        if >0 and <100, clip pixels with diagonal-normalized values
        higher than the specified percentile of matrix-wide values.
    sort_metric : str
        If provided, re-sort `eigenvecs` and `eigvals` in the order of
        decreasing correlation between phasing_track and eigenvector, using the
        specified measure of correlation. Possible values:
        'pearsonr' - sort by decreasing Pearson correlation.
        'var_explained' - sort by decreasing absolute amount of variation in
        `eigvecs` explained by `phasing_track` (i.e. R^2 * var(eigvec))
        'MAD_explained' - sort by decreasing absolute amount of Median Absolute
        Deviation from the median of `eigvecs` explained by `phasing_track`
        (i.e. COMED(eigvec, phasing_track) * MAD(eigvec)).
        'spearmanr' - sort by decreasing Spearman correlation.
        This option is designed to report the most "biologically" informative
        eigenvectors first, and prevent eigenvector swapping caused by
        translocations. In reality, however, sometimes it shows poor
        performance and may lead to reporting of non-informative eigenvectors.
        Off by default.


    Returns
    -------
    eigenvalues, eigenvectors

    .. note:: ALWAYS check your EVs by eye. The first one occasionally does
              not reflect the compartment structure, but instead describes
              chromosomal arms or translocation blowouts.

    """
    A = np.array(A)
    A[~np.isfinite(A)] = 0

    mask = A.sum(axis=0) > 0

    if A.shape[0] <= ignore_diags + 3 or mask.sum() <= ignore_diags + 3:
        return (
            np.array([np.nan for i in range(n_eigs)]),
            np.array([np.ones(A.shape[0]) * np.nan for i in range(n_eigs)]),
        )

    if ignore_diags:
        for d in range(-ignore_diags + 1, ignore_diags):
            numutils.set_diag(A, 1.0, d)

    OE, _, _, _ = numutils.observed_over_expected(A, mask)

    if clip_percentile and clip_percentile < 100:
        OE = np.clip(OE, 0, np.percentile(OE[mask, :][:, mask], clip_percentile))

    # subtract 1.0
    OE -= 1.0

    # empty invalid rows, so that get_eig can find them
    OE[~mask, :] = 0
    OE[:, ~mask] = 0

    eigvecs, eigvals = numutils.get_eig(OE, n_eigs, mask_zero_rows=True)
    eigvecs /= np.sqrt(np.nansum(eigvecs ** 2, axis=1))[:, None]
    eigvecs *= np.sqrt(np.abs(eigvals))[:, None]

    # Orient and reorder
    if phasing_track is not None:
        eigvals, eigvecs = _phase_eigs(eigvals, eigvecs, phasing_track, sort_metric)

    return eigvals, eigvecs


def _filter_heatmap(A, transmask, perc_top, perc_bottom):
    # Truncate trans blowouts
    lim = np.percentile(A[transmask], perc_top)
    tdata = A[transmask]
    tdata[tdata > lim] = lim
    A[transmask] = tdata

    # Remove bins with poor coverage in trans
    marg = np.sum(A, axis=0)
    marg_nz = marg[np.sum(A, axis=0) > 0]
    min_cutoff = np.percentile(marg_nz, perc_bottom)
    dropmask = (marg > 0) & (marg < min_cutoff)
    A[dropmask, :] = 0
    A[:, dropmask] = 0
    return A


def _fake_cis(A, cismask):
    cismask = cismask.astype(np.uint8)
    s = np.abs(np.sum(A, axis=0)) <= 1e-10
    cismask[:, s] = 2
    cismask[s, :] = 2
    numutils.fake_cis(A, cismask)
    return A


def trans_eig(
    A,
    partition,
    n_eigs=3,
    perc_top=99.95,
    perc_bottom=1,
    phasing_track=None,
    sort_metric=False,
):
    """
    Compute compartmentalization eigenvectors on trans contact data

    Parameters
    ----------
    A : 2D array
        balanced whole genome contact matrix
    partition : sequence of int
        bin offset of each contiguous region to treat separately (e.g.,
        chromosomes or chromosome arms)
    n_eigs : int
        number of eigenvectors to compute; default = 3
    perc_top : float (percentile)
        filter - clip trans blowout contacts above this cutoff; default = 99.95
    perc_bottom : float (percentile)
        filter - remove bins with trans coverage below this cutoff; default=1
    phasing_track : 1D array, optional
        if provided, eigenvectors are flipped to achieve a positive correlation
        with `phasing_track`.
    sort_metric : str
        If provided, re-sort `eigenvecs` and `eigvals` in the order of
        decreasing correlation between phasing_track and eigenvector, using the
        specified measure of correlation. Possible values:
        'pearsonr' - sort by decreasing Pearson correlation.
        'var_explained' - sort by decreasing absolute amount of variation in
        `eigvecs` explained by `phasing_track` (i.e. R^2 * var(eigvec))
        'MAD_explained' - sort by decreasing absolute amount of Median Absolute
        Deviation from the median of `eigvecs` explained by `phasing_track`
        (i.e. COMED(eigvec, phasing_track) * MAD(eigvec)).
        'spearmanr' - sort by decreasing Spearman correlation.
        This option is designed to report the most "biologically" informative
        eigenvectors first, and prevent eigenvector swapping caused by
        translocations. In reality, however, sometimes it shows poor
        performance and may lead to reporting of non-informative eigenvectors.
        Off by default.


    Returns
    -------
    eigenvalues, eigenvectors

    .. note:: ALWAYS check your EVs by eye. The first one occasionally does
          not reflect the compartment structure, but instead describes
          chromosomal arms or translocation blowouts.


    """
    A = np.array(A)
    if A.shape[0] != A.shape[1]:
        raise ValueError("A is not symmetric")

    n_bins = A.shape[0]
    if not (
        partition[0] == 0 and partition[-1] == n_bins and np.all(np.diff(partition) > 0)
    ):
        raise ValueError(
            "Not a valid partition. Must be a monotonic sequence "
            "from 0 to {}.".format(n_bins)
        )

    # Delete cis data and create trans mask
    extents = zip(partition[:-1], partition[1:])
    part_ids = []
    for n, (i0, i1) in enumerate(extents):
        A[i0:i1, i0:i1] = 0
        part_ids.extend([n] * (i1 - i0))
    part_ids = np.array(part_ids)
    is_trans = part_ids[:, None] != part_ids[None, :]

    # Filter heatmap
    is_bad_bin = np.nansum(A, axis=0) == 0
    is_good_bin = ~is_bad_bin
    is_valid = np.logical_and.outer(is_good_bin, is_good_bin)
    A[is_bad_bin, :] = 0
    A[:, is_bad_bin] = 0

    A = _filter_heatmap(A, is_trans & is_valid, perc_top, perc_bottom)
    is_bad_bin = np.nansum(A, axis=0) == 0
    is_good_bin = ~is_bad_bin
    is_valid = np.logical_and.outer(is_good_bin, is_good_bin)
    A[is_bad_bin, :] = 0
    A[:, is_bad_bin] = 0

    # Fake cis and re-balance
    A = _fake_cis(A, ~is_trans)
    A = numutils.iterative_correction_symmetric(A)[0]
    A = _fake_cis(A, ~is_trans)
    A = numutils.iterative_correction_symmetric(A)[0]

    # Compute eig
    Abar = np.mean(A[is_valid])  # center by scalar mean
    O = (A - Abar) / Abar
    O[is_bad_bin, :] = 0
    O[:, is_bad_bin] = 0
    eigvecs, eigvals = numutils.get_eig(O, n_eigs, mask_zero_rows=True)

    eigvecs /= np.sqrt(np.nansum(eigvecs ** 2, axis=1))[:, None]
    eigvecs *= np.sqrt(np.abs(eigvals))[:, None]
    if phasing_track is not None:
        eigvals, eigvecs = _phase_eigs(eigvals, eigvecs, phasing_track, sort_metric)

    return eigvals, eigvecs


def cooler_cis_eig(
    clr,
    bins,
    regions=None,
    n_eigs=3,
    phasing_track_col="GC",
    balance="weight",
    ignore_diags=None,
    bad_bins=None,
    clip_percentile=99.9,
    sort_metric=None,
    map=map,
):
    """
    Compute compartment eigenvector for a given cooler `clr` in a number of
    symmetric intra chromosomal regions (cis-regions), or for each chromosome.
    Note that the amplitude of compartment eigenvectors is weighted by their
    corresponding eigenvalue
    Parameters
    ----------
    clr : cooler
        cooler object to fetch data from
    bins : DataFrame
        table of bins derived from clr with phasing track added
    regions : iterable or DataFrame, optional
        if provided, eigenvectors are calculated for the regions only,
        otherwise chromosome-wide eigenvectors are computed, for chromosomes
        specified in bins.
    n_eigs : int
        number of eigenvectors to compute
    phasing_track_col : str, optional
        name of the columns in `bins` table, if provided, eigenvectors are
        flipped to achieve a positive correlation with `bins[phasing_track_col]`.
    balance : str
        name of the column with balancing weights to be used.
    ignore_diags : int, optional
        the number of diagonals to ignore. Derived from cooler metadata
        if not specified.
    bad_bins : array-like
        a list of bins to ignore. Indexes of bins must be absolute,
        as in clr.bins()[:], as opposed to being offset by chromosome start.
        `bad_bins` will be combined with the bad bins masked by balancing.
    clip_percentile : float
        if >0 and <100, clip pixels with diagonal-normalized values
        higher than the specified percentile of matrix-wide values.
    sort_metric : str
        If provided, re-sort `eigenvecs` and `eigvals` in the order of
        decreasing correlation between phasing_track and eigenvector, using the
        specified measure of correlation. Possible values:
        'pearsonr' - sort by decreasing Pearson correlation.
        'var_explained' - sort by decreasing absolute amount of variation in
        `eigvecs` explained by `phasing_track` (i.e. R^2 * var(eigvec))
        'MAD_explained' - sort by decreasing absolute amount of Median Absolute
        Deviation from the median of `eigvecs` explained by `phasing_track`
        (i.e. COMED(eigvec, phasing_track) * MAD(eigvec)).
        'spearmanr' - sort by decreasing Spearman correlation.
        This option is designed to report the most "biologically" informative
        eigenvectors first, and prevent eigenvector swapping caused by
        translocations. In reality, however, sometimes it shows poor
        performance and may lead to reporting of non-informative eigenvectors.
        Off by default.
    map : callable, optional
        Map functor implementation.
    Returns
    -------
    eigvals, eigvec_table -> DataFrames with eigenvalues for each region and
    a table of eigenvectors filled in the `bins` table.
    .. note:: ALWAYS check your EVs by eye. The first one occasionally does
              not reflect the compartment structure, but instead describes
              chromosomal arms or translocation blowouts. Possible mitigations:
              employ `regions` (e.g. arms) to avoid issues with chromosomal arms,
              use `bad_bins` to ignore small transolcations.
    """

    # get chromosomes from bins, if regions not specified:
    if regions is None:
        regions = list(bins["chrom"].unique()) # parse_regions fill in the rest

    # make sure phasing_track_col is in bins, if phasing is requested
    if phasing_track_col and (phasing_track_col not in bins):
        raise ValueError(f'No column "{phasing_track_col}" in the bin table')

    # regions to dataframe
    regions = bioframe.parse_regions(regions, clr.chromsizes)

    # ignore diags as in cooler inless specified
    ignore_diags = (
        clr._load_attrs("bins/weight").get("ignore_diags", 2)
        if ignore_diags is None
        else ignore_diags
    )

    # prepare output table for eigen vectors
    eigvec_table = bins.copy()
    eigvec_columns = [f"E{i + 1}" for i in range(n_eigs)]
    for ev_col in eigvec_columns:
        eigvec_table[ev_col] = np.nan

    # prepare output table for eigenvalues
    eigvals_table = regions.copy()
    eigval_columns = [f"eigval{i + 1}" for i in range(n_eigs)]
    for eval_col in eigval_columns:
        eigvals_table[eval_col] = np.nan

    def _each(region):
        """
        perform eigen decomposition for a given region
        assuming safety checks are done outside of this
        function.
        Parameters
        ----------
        region: tuple-like
            tuple of the form (chroms,start,end,*)
        Returns
        -------
        _region, eigvals, eigvecs -> ndarrays
            array of eigenvalues and an array eigenvectors
        """
        _region = region[:3] # take only (chrom, start, end)
        A = clr.matrix(balance=balance).fetch(_region)

        # filter bad_bins relevant for the _region from A
        if bad_bins is not None:
            # filter bad_bins for the _region and turn relative:
            lo, hi = clr.extent(_region)
            bad_bins_region = bad_bins[(bad_bins>=lo)&(bad_bins<hi)]
            bad_bins_region -= lo
            if len(bad_bins_region) > 0:
                # apply bad bins to symmetric matrix A:
                A[:,bad_bins_region] = np.nan
                A[bad_bins_region,:] = np.nan

        # extract phasing track relevant for the _region
        phasing_track = (
            bioframe.select(bins, _region)[phasing_track_col].values
            if phasing_track_col
            else None
        )

        eigvals, eigvecs = cis_eig(
            A,
            n_eigs=n_eigs,
            ignore_diags=ignore_diags,
            phasing_track=phasing_track,
            clip_percentile=clip_percentile,
            sort_metric=sort_metric,
        )

        return _region, eigvals, eigvecs

    # eigendecompose matrix per region (can be multiprocessed)
    # output assumes that the order of results matches regions
    results = map(_each, regions.values)

    # go through eigendecomposition results and fill in
    # output table eigvec_table and eigvals_table
    for _region, _eigvals, _eigvecs in results:
        idx = bioframe.select(eigvec_table, _region).index
        eigvec_table.at[idx, eigvec_columns] = _eigvecs.T
        idx = bioframe.select(eigvals_table, _region).index
        eigvals_table.at[idx, eigval_columns] = _eigvals


    return eigvals_table, eigvec_table


def cooler_trans_eig(
    clr,
    bins,
    n_eigs=3,
    partition=None,
    phasing_track_col="GC",
    balance="weight",
    sort_metric=None,
    **kwargs
):

    if partition is None:
        partition = np.r_[
            [clr.offset(chrom) for chrom in clr.chromnames], len(clr.bins())
        ]

    lo = partition[0]
    hi = partition[-1]
    A = clr.matrix(balance=balance)[lo:hi, lo:hi]
    bins = bins[lo:hi]

    phasing_track = None
    if phasing_track_col:
        if phasing_track_col not in bins:
            raise ValueError(
                'No column "{}" in the bin table'.format(phasing_track_col)
            )
        phasing_track = bins[phasing_track_col].values[lo:hi]

    eigvals, eigvecs = trans_eig(
        A,
        partition,
        n_eigs=n_eigs,
        phasing_track=phasing_track,
        sort_metric=sort_metric,
        **kwargs
    )

    eigvec_table = bins.copy()
    for i, eigvec in enumerate(eigvecs):
        eigvec_table["E{}".format(i + 1)] = eigvec

    eigvals = pd.DataFrame(
        data=np.atleast_2d(eigvals),
        columns=["eigval" + str(i + 1) for i in range(n_eigs)],
    )
    return eigvals, eigvec_table



import os
import io
import gzip
import tarfile
import tempfile

from . import fastsavetxt

import cooler


def dump_cworld(
    in_cooler,
    out=None,
    region=None,
    iced=False,
    iced_unity=False,
    buffer_size=int(1e8),
):
    """
    Dump a genome-wide contact matrix from cooler into a CWorld-format
    text matrix.

    Parameters
    ----------
    in_cooler : str or cooler
        A cooler object or the path to the file.

    out : str or file object
        Either:
        -- a path to the output file. If ends with .gz the output is gzipped
        -- a file object
        -- a stdin of a Popen object
        -- None, in which case the data is dumped into a string and returned
        TIP: when using files/stdin do not forget to flush()/communicate().

    region : str
        The region to dump. By default is None, dump the genome-wide matrix.

    iced : bool, optional
        If True, dump the balanced matrix.

    iced_unity : bool, optional
        If True and `iced` is True, dump the matrix balanced to a unity.

    buffer_size : int
        The chunk size for iterating over the rows of the Hi-C matrix.
    """

    # Prepare the out pipe and the clean-up function.
    if not (out):
        out = io.BytesIO(b"")
    if issubclass(type(out), str) or issubclass(type(out), bytearray):
        if out.endswith(".gz"):
            writer = fastsavetxt.gzipWriter(out)
            out_pipe = writer.stdin
            close_out_func = writer.communicate
        else:
            writer = open(out, "wb")
            out_pipe = writer
            close_out_func = writer.flush
    elif hasattr(out, "write"):
        out_pipe = out
        close_out_func = fastsavetxt.empty_func

    # Make headers
    if not issubclass(type(in_cooler), cooler.Cooler):
        c = cooler.Cooler(in_cooler)
    else:
        c = in_cooler

    res = c.info["bin-size"]
    gname = c.info["genome-assembly"]

    bins = c.bins()[:] if not (region) else c.bins().fetch(region)
    nbins = len(bins)

    col_headers = "\t".join(
        ["{}x{}".format(nbins, nbins)]
        + [
            "{}|{}|{}:{}-{}".format(binidx, gname, b.chrom, b.start + 1, b.end)
            for binidx, b in bins.iterrows()
        ]
    ).encode()

    row_headers = [
        "{}|{}|{}:{}-{}".format(binidx1, gname, b1.chrom, b1.start + 1, b1.end).encode()
        for binidx1, b1 in bins.iterrows()
    ]

    # Iterate over a matrix one block at a time.
    nrows_per_step = max(1, buffer_size // nbins)
    for i in range(nbins // nrows_per_step + 1):
        lo = min(nbins, i * nrows_per_step)
        hi = min(nbins, (i + 1) * nrows_per_step)
        if hi <= lo:
            break
        mat = (
            c.matrix(balance=iced)
            if not (region)
            else c.matrix(balance=iced).fetch(region)
        )[lo:hi]
        if iced and (not iced_unity):
            mat *= c._load_attrs("/bins/weight")["scale"]

        fastsavetxt.array2txt(
            mat,
            out_pipe,
            format_string=b"%.8f" if iced_unity else b"%.4lf",
            header=col_headers if i == 0 else None,
            row_headers=row_headers[lo:hi],
        )

    if issubclass(type(out), io.BytesIO):
        return out.getvalue()
    else:
        close_out_func()


def dump_cworld_tar(
    cooler_paths, out_path,
):
    """
    Makes a CWorld .tar archive with binned contact maps at multiple resolutions
    in .matrix.txt.gz format.

    Parameters
    ----------
    cooler_paths : a list of str
        The paths to all coolers to dump into a single CWorld tar archive.
        Must correspond to the same dataset and have different resolutions.

    out_path : str
        The path to the output file.

    """

    dataset_name = os.path.splitext(os.path.split(out_path)[1])[0]

    with tempfile.TemporaryDirectory() as cworld_tmp_path:
        for cooler_path in cooler_paths:
            res = cooler.Cooler(cooler_path).info["bin-size"]
            os.mkdir(os.path.join(cworld_tmp_path, "C-" + str(res)))
            for iced, iced_label in [(True, "iced"), (False, "raw")]:
                folder_path = os.path.join(cworld_tmp_path, "C-" + str(res), iced_label)
                os.mkdir(folder_path)

                mat_path = os.path.join(
                    folder_path,
                    "{}__C-{}-{}.matrix.gz".format(dataset_name, res, iced_label),
                )

                dump_cworld(
                    in_cooler=cooler_path, out=mat_path, iced=iced, iced_unity=False
                )

        with tarfile.open(out_path, mode="w") as archive:
            archive.add(cworld_tmp_path, arcname=dataset_name, recursive=True)
